from datetime import time
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.support.wait import WebDriverWait

from selenium_super.enums import Until, ElementActions
import numpy as np
import scipy.interpolate as si


class Actions:
    def __init__(self, driver):
        self.driver = driver
        self.scroll = Scroll(self.driver)
        self.mouse_events = MouseEvents(self.driver)
        self.wait = Wait(self.driver)

    def focus(self, element: WebElement) -> None:
        """ Focuses on an element. """
        self.driver.execute_script('arguments[0].focus();', element)

    def click(self, element: WebElement) -> None:
        """ Clicks on an element. """
        self.driver.execute_script('arguments[0].click();', element)

    def blur(self, element: WebElement) -> None:
        """ Clear the focus from a selected web element. """
        self.driver.execute_script('arguments[0].blur();', element)

    def scroll_into_view(self, element: WebElement) -> None:
        """ Scrolls the element into view.  """
        self.driver.execute_script("arguments[0].scrollIntoView();", element)

    def get_bounding_client_rect(self, element: WebElement) -> dict:
        """ Gets the bounding client rect of an element. """
        return self.driver.execute_script('return arguments[0].getBoundingClientRect();', element)

    def get_attribute(self, element: WebElement, attribute: str):
        """ Gets an element's attribute. """
        return self.driver.execute_script('return arguments[0].getAttribute(arguments[1]);', element, attribute)

    def set_attribute(self, element: WebElement, attribute: str, value):
        """ Sets an element's attribute. """
        return self.driver.execute_script(
            'return arguments[0].setAttribute(arguments[1], arguments[2]);', element, attribute, value
        )

    def remove_attribute(self, element: WebElement, attribute: str):
        """ Removes an element's attribute. """
        return self.driver.execute_script('return arguments[0].removeAttribute(arguments[1]);', element, attribute)

    def get_timezone_offset(self) -> int:
        """ Gets the timezone offset of the browser in minutes. """
        return self.execute_script('return new Date().getTimezoneOffset();')

    def set_coordinates(self, coordinates: tuple) -> None:
        """ Sets the geolocation for location services. """
        script = '''var latitude = arguments[0];
                    var longitude = arguments[1];
                    window.navigator.geolocation.getCurrentPosition = function(success) {
                        var position = {
                            "coords" : {
                                "latitude": latitude,
                                "longitude": longitude
                            }
                        };
                        success(position);
                }'''
        self.driver.execute_script(
            script,
            *coordinates,
        )

    def get_coordinates(self) -> tuple:
        script = '''var latitude = ''
                    var longitude = ''
                    window.navigator.geolocation.getCurrentPosition(function(pos) {
                        latitude = pos.coords.latitude;
                        longitude = pos.coords.longitude;
                    });
                    return [latitude, longitude];'''
        return self.driver.execute_script(script)


class Scroll:
    def __init__(self, driver):
        self.driver = driver

    def scroll_to_bottom(self, scroll_steps=250, bottom_crop_height=1000):
        verical_ordinate = scroll_steps
        page_height = self.driver.execute_script(
            "return Math.max(document.body.scrollHeight, document.body.offsetHeight, document.documentElement.clientHeight, document.documentElement.scrollHeight, document.documentElement.offsetHeight)"
        )
        total_jumps = int((page_height - bottom_crop_height) / scroll_steps)
        for i in range(0, total_jumps):
            self.driver.execute_script(
                "window.scrollTo(window.scrollY, {});".format(
                    verical_ordinate))
            verical_ordinate += scroll_steps
            time.sleep(0.20)

    def infinite_scroll_to_bottom(self, scroll_steps=250):
        verical_ordinate = scroll_steps
        last_height = None

        while True:
            self.driver.execute_script(f"window.scrollTo(window.scrollY, {verical_ordinate});")
            verical_ordinate += scroll_steps
            height = self.driver.execute_script("return (window.innerHeight + window.scrollY)")
            if last_height != height:
                last_height = height
                time.sleep(0.20)
            else:
                break

    def infinite(self, loading_selector, xpath):
        # Get scroll height after first time page load
        last_height = self.driver.execute_script("return document.body.scrollHeight")
        # scroll_element = self.driver.execute_script(f'return document.evaluate("{xpath}", document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue')
        scroll_element = f'document.evaluate("{xpath}", document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue'
        while scroll_element:
            # Scroll down to bottom
            self.driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
            # Wait until element dissapear
            self.wait_until_element_disappear(loading_selector)
            print(last_height)
            # Calculate new scroll height and compare with last scroll height
            # new_height = self.driver.execute_script("return document.body.scrollHeight")
            elementProps = self.driver.execute_script(f'return {scroll_element}.getBoundingClientRect()')
            print(elementProps)
            new_height = self.driver.execute_script(
                f"return ({elementProps.get('height')} + {scroll_element}.scrollTop)")

            if new_height == last_height:
                break
            last_height = new_height

    def infinite_scroll_to_bottom_by_xpath(self, xpath=None, scroll_steps=250, data_append_function=None,
                                           loading_selector=None):
        try:
            data = []

            verical_ordinate = scroll_steps
            last_height = None
            scroll_element = f'document.evaluate("{xpath}", document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue'
            # print(scroll_element)
            while scroll_element:
                print(f"{scroll_element}.scrollTo({scroll_element}.scrollTop, {verical_ordinate});")

                verical_ordinate += scroll_steps

                self.driver.execute_script(
                    f"{scroll_element}.scrollTo({scroll_element}.scrollTop, {verical_ordinate});")
                # self.driver.execute_script(f"window.scrollTo(0, {verical_ordinate});")
                if loading_selector:
                    self.wait_until_element_disappear(loading_selector)

                elementProps = self.driver.execute_script(f'return {scroll_element}.getBoundingClientRect()')
                height = self.driver.execute_script(
                    f"return ({elementProps.get('height')} + {scroll_element}.scrollTop)") if elementProps else 500

                if last_height != height:
                    last_height = height
                    if data_append_function:
                        data = data + (data_append_function() or [])
                    else:
                        time.sleep(0.20)
                else:
                    break

            if data_append_function: return list(data)
        except Exception as e:
            print(f'Error while trying to do infinite scroll on XPATH - {xpath}\nError: {e}')


class Wait:
    def __init__(self, driver):
        self.driver = driver

    def until(self, wait_element, method: Until, wait: int, base_element=None, action: ElementActions = None):
        base_element = base_element or self.driver
        element = WebDriverWait(base_element, wait).until(method(wait_element))
        if action and element:
            action(element)
        return element

    def until_not(self, wait_element, method: Until, wait: int, base_element=None, action: ElementActions = None):
        base_element = base_element or self.driver
        element = WebDriverWait(base_element, wait).until_not(method(wait_element))
        if action and element:
            action(element)
        return element

    def until_clickable_element(self, wait_element, wait, base_element=None, click=True):
        self.until(wait_element, Until.clickable, wait, base_element, action=ElementActions.click if click else None)

    def until_element_invisible(self, wait_element, wait, base_element=None):
        self.driver.implicitly_wait(0)
        return self.until(wait_element, Until.invisible, wait, base_element)


class MouseEvents:
    def __init__(self, driver):
        self.driver = driver

    def _trigger_mouse_event(self, element, click_event):
        script = '''var clickEvent = document.createEvent('MouseEvents');
                    clickEvent.initEvent(arguments[1], true, true);
                    arguments[0].dispatchEvent(clickEvent);'''
        self.driver.execute_script(script, element, click_event)

    def click(self, element):
        self._trigger_mouse_event(element, 'click')

    def mouse_up(self, element):
        self._trigger_mouse_event(element, 'mouseup')

    def mouse_down(self, element):
        self._trigger_mouse_event(element, 'mousedown')

    def mouse_over(self, element):
        self._trigger_mouse_event(element, 'mouseover')

    def human_like_mouse_move(self, base_element, click=False):
        action = ActionChains(self.driver)

        points = [[6, 2], [3, 2], [0, 0], [0, 2]]
        points = np.array(points)
        x = points[:, 0]
        y = points[:, 1]

        t = range(len(points))
        ipl_t = np.linspace(0.0, len(points) - 1, 100)

        x_tup = si.splrep(t, x, k=1)
        y_tup = si.splrep(t, y, k=1)

        x_list = list(x_tup)
        xl = x.tolist()
        x_list[1] = xl + [0.0, 0.0, 0.0, 0.0]

        y_list = list(y_tup)
        yl = y.tolist()
        y_list[1] = yl + [0.0, 0.0, 0.0, 0.0]

        x_i = si.splev(ipl_t, x_list)
        y_i = si.splev(ipl_t, y_list)

        action.move_to_element(base_element)

        if click:
            action.click().perform()
        else:
            action.perform()

        c = 5
        i = 0
        for mouse_x, mouse_y in zip(x_i, y_i):
            action.move_by_offset(mouse_x, mouse_y)
            action.perform()
            i += 1
            if i == c:
                break



class Other:
    def __init__(self, driver):
        self.driver = driver




from .selenium_super import SuperWebDriver
from .telegram_api import TelegramApi
from .telethon_helpers import TelethonHelpers
from .file_utils import FileUtils
from .folder_utils import FolderUtils
from .time_utils import TimeUtils
from .nordvpn import NordVPN

__all__ = (
    'SuperWebDriver',
    'TelegramApi',
    'TelethonHelpers',
    'FileUtils',
    'FolderUtils',
    'TimeUtils',
    'NordVPN'
)

VERSION = '0.0.1'


import platform
from selenium.webdriver.firefox.options import Options
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium import webdriver
from selenium_super.actions import Actions

from selenium_super.js import Js
import asyncio
import base64
import json
import codecs

from .process_multi_tasks import MultiTask, ProcessMultiTasks
from .auth import Auth
from .enums import By, Until, ElementActions
from .file_utils import FileUtils
from .folder_utils import FolderUtils
from .helpers import get_args_for_elements_by_selector
from .time_utils import TimeUtils
from .nordvpn import NordVPN


class SuperWebDriver(webdriver.Firefox, webdriver.Chrome):
    def __init__(self, config_path, selectors_path=None):

        self.default_firefox_options = None
        self.default_chrome_options = None
        self.default_linux_options = None
        self.firefox_binary_location = None
        self.firefox_executable_path = None
        self.chrome_logger = None
        self.chrome_extensions = None
        self.chrome_profile = None
        self.chrome_executable_path = None
        self.download_folder = None
        self.headless = None
        self.options = None
        self.browser = None
        self.profile = None
        self.vpn_countries = None
        self.use_vpn = None
        self.selectors = {}
        ## load utils
        self.file_utils = FileUtils()
        self.folder_utils = FolderUtils()
        self.time_utils = TimeUtils()
        self.js = Js(self)
        self.auth = Auth(self)
        self.actions = Actions(self)
        self.vpn = NordVPN()

        self.load_browser_config(config_path)
        if self.use_vpn and self.vpn_countries:
            self.vpn.set_countries(self.vpn_countries)
            self.vpn.rotate(google_check=True)

        self.get_web_driver(init=True)

        if selectors_path is not None:
            self.load_selectors_file(selectors_path)

    def load_browser_config(self, file_path):
        config = self.file_utils.load_json_file(file_path=file_path)
        if not config:
            print('Could not load browser config file!')
        else:
            self.browser = config.get('use_browser', None)
            self.options = config.get('options', None)
            self.headless = config.get('headless', None)
            self.download_folder = config.get('download_folder', None)
            self.use_vpn = config.get('use_vpn', False)
            self.vpn_countries = config.get('vpn_countries', [])

            self.chrome_executable_path = config.get('chrome_executable_path', None)
            self.chrome_profile = config.get('chrome_profile', None)
            self.chrome_logger = config.get('chrome_logger', None)
            self.chrome_extensions = config.get('extensions', [])

            self.firefox_executable_path = config.get('firefox_executable_path', None)
            self.firefox_binary_location = config.get('firefox_binary_location', None)

            self.default_linux_options = ['--no-sandbox', '--disable-gpu', '--disable-dev-shm-usage',
                                          '--profile-directory=Default',
                                          '--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.101 Safari/537.36',
                                          '--user-data-dir=~/.config/google-chrome']
            self.default_chrome_options = ['--log-level=3', 'window-size=1920,1080',
                                           '--user-agent=Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.5112.101 Safari/537.36']
            self.default_firefox_options = []

    def get_web_driver(self, init=False) -> webdriver:
        if init is False and self.driver is not None:
            return self.driver
        elif platform.system() == 'Linux':
            print('Starting selenium using Chrome - Linux')
            options = ChromeOptions()
            if self.headless is not None:
                options.headless = self.headless
            for option in self.options or self.default_linux_options:
                options.add_argument(option)
            webdriver.Chrome.__init__(self, options=options)
        elif self.browser == 'chrome':
            print('Starting selenium using Chrome - Windows')
            capabilities = DesiredCapabilities.CHROME.copy()
            options = ChromeOptions()
            for option in self.options or self.default_chrome_options:
                options.add_argument(option)

            for extension in self.chrome_extensions:
                options.add_extension(extension)

            if self.chrome_logger:
                capabilities['goog:loggingPrefs'] = {'performance': 'ALL'}
            if self.headless is not None:
                options.headless = self.headless
            if self.chrome_profile:
                options.add_argument(f"--user-data-dir={self.chrome_profile}")
            if self.download_folder:
                options.add_experimental_option(
                    "prefs", {
                        "download.default_directory":
                            self.download_folder,
                        "download.prompt_for_download": False,
                        "download.directory_upgrade": True,
                        "safebrowsing.enabled": True
                    })

            webdriver.Chrome.__init__(self,
                                      executable_path=self.chrome_executable_path,
                                      options=options, desired_capabilities=capabilities)
        elif self.browser == 'firefox':
            print('Starting selenium using Chrome - Firefox')
            options = Options()
            options.binary_location = self.firefox_binary_location
            for option in self.options or self.default_firefox_options:
                options.add_argument(option)

            options.profile = []
            webdriver.Firefox.__init__(self,
                                       executable_path=self.firefox_executable_path,
                                       options=options)
        else:
            print('Could not start driver, please fix browser_config.json file')

        self.driver = self

    def validate_element(self, element):
        return element is not None \
               and element.get('by') is not None \
               and element.get('value') is not None

    def load_selectors_file(self, path):
        config = FileUtils().load_json_file(path)
        by_options = {
            "By.XPATH": By.XPATH,
            "By.NAME": By.NAME,
            "By.CLASS_NAME": By.CLASS_NAME,
            "By.CSS_SELECTOR": By.CSS_SELECTOR,
            "By.ID": By.ID,
            "By.LINK_TEXT": By.LINK_TEXT,
            "By.TAG_NAME": By.TAG_NAME
        }

        loaded_config = {}
        for key, val in config.items():
            by = by_options[val.get('by')]
            value = val.get('value')
            loaded_config[key] = {"by": by, "value": value}

        self.selectors = loaded_config
        return loaded_config

    def get_selectors(self):
        return self.selectors

    def get_selector_value(self, selector):
        return self.selectors.get(selector, {}).get('value', None)

    def get_elements_by_selector(self,
                                 selector_name=None,
                                 base_element=None,
                                 multiple=False,
                                 attribute=None,
                                 filter=None,
                                 xpath=None,
                                 wait=0, action: ElementActions = None):

        try:
            if xpath is None:
                by = self.selectors.get(selector_name)
                value = self.selectors.get(selector_name)

                if by is None or value is None:
                    print('No selector {} found in configuration file'.format(
                        selector_name))
                    return None

                by = by.get('by', None)
                value = value.get('value', None)

            else:
                by = By.XPATH
                value = xpath

            search_element = self.driver if base_element is None else base_element

            if not multiple:
                element = self.actions.wait.until(wait_element=(by, value), wait=wait, base_element=search_element, method=Until.present)

                if action and element:
                    action(element)

                if attribute is not None and element is not None:
                    return element.get_attribute(
                        attribute) if filter is None else filter(
                        element.get_attribute(attribute))
                return element

            elif multiple:
                elements = self.actions.wait.until(wait_element=(by, value), wait=wait, base_element=search_element,
                                                  method=Until.presents)

                if attribute is not None and elements is not None:
                    attributes_data = []
                    for element in elements:
                        if action and element:
                            action(element)
                        if filter is None:
                            attributes_data.append(
                                element.get_attribute(attribute))
                        else:
                            attributes_data.append(
                                filter(element.get_attribute(attribute)))
                    return attributes_data
                return elements
        except Exception as e:
            print(e)
            return None


    def get_multiple_elements_by_selectors(self, selectors, return_when=asyncio.ALL_COMPLETED):
        processor = ProcessMultiTasks()
        data = processor.run(tasks=[
            MultiTask(self.get_elements_by_selector, selector_args[0], args=(selector_args)) for selector_args in
            [get_args_for_elements_by_selector(selector) for selector in selectors]
        ], return_when=return_when)

        return data

    def get_file_content_chrome(self, uri):
        result = self.driver.execute_async_script("""
            var uri = arguments[0];
            var callback = arguments[1];
            var toBase64 = function(buffer){for(var r,n=new Uint8Array(buffer),t=n.length,a=new Uint8Array(4*Math.ceil(t/3)),i=new Uint8Array(64),o=0,c=0;64>c;++c)i[c]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".charCodeAt(c);for(c=0;t-t%3>c;c+=3,o+=4)r=n[c]<<16|n[c+1]<<8|n[c+2],a[o]=i[r>>18],a[o+1]=i[r>>12&63],a[o+2]=i[r>>6&63],a[o+3]=i[63&r];return t%3===1?(r=n[t-1],a[o]=i[r>>2],a[o+1]=i[r<<4&63],a[o+2]=61,a[o+3]=61):t%3===2&&(r=(n[t-2]<<8)+n[t-1],a[o]=i[r>>10],a[o+1]=i[r>>4&63],a[o+2]=i[r<<2&63],a[o+3]=61),new TextDecoder("ascii").decode(a)};
            var xhr = new XMLHttpRequest();
            xhr.responseType = 'arraybuffer';
            xhr.onload = function(){ callback(toBase64(xhr.response)) };
            xhr.onerror = function(){ callback(xhr.status) };
            xhr.open('GET', uri);
            xhr.send();
            """, uri)
        if type(result) == int:
            raise Exception("Request failed with status %s" % result)
        return base64.b64decode(result)

    def save_screenshot(self, name, path, element):
        full_path = f"{path}{name}.png"
        print("Saving image in - {}".format(full_path))
        element.screenshot(full_path)
        return full_path

    def save_html(self, name, path):
        with codecs.open(f"{path}{name}.html", "w", "utf-8") as html:
            full_path = f"{path}{name}.html"
            print("Saving HTML in - {}".format(full_path))
            html.write(self.driver.page_source)
            return full_path

    def record_responses(self):
        logs = self.driver.get_log('performance')
        responses = []
        for log in logs:
            if log['message']:
                log = json.loads(log['message'])['message']
                try:
                    if ("Network.responseReceived" in log["method"] and "json" in log["params"]["response"][
                        "mimeType"]):
                        try:
                            body = self.driver.execute_cdp_cmd('Network.getResponseBody',
                                                               {'requestId': log["params"]["requestId"]})
                            log['body'] = json.loads(body['body'])
                            responses.append(log)
                        except Exception as e:
                            responses.append(log)
                except Exception as e:
                    print(e)

        return responses

    def save_responses(self, file_path=None):
        logs = self.record_responses()
        json_object = {
            'logs': logs,
        }
        if file_path: file = FileUtils().write_json_file(file_path=file_path, write_mode='w+', json_object=json_object,
                                                         stringify=True, log=True)
        return logs

    def click_element_by_xpath(self, xpath):
        self.execute_script(
            """document.evaluate({}, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue?.click()""".format(
                xpath))

    def send_devtools(self, cmd, params={}):
        resource = f"/session/{self.driver.session_id}/chromium/send_command_and_get_result"
        url = self.driver.command_executor._url + resource
        body = json.dumps({'cmd': cmd, 'params': params})
        response = self.driver.command_executor._request('POST', url, body)
        if response.get('status'):
            raise Exception(response.get('value'))
        return response.get('value')

    def get_pdf_from_html(self, print_options={}, save_path=None):
        calculated_print_options = {
            'landscape': False,
            'displayHeaderFooter': False,
            'printBackground': True,
            'preferCSSPageSize': True,
        }
        calculated_print_options.update(print_options)
        result = self.send_devtools("Page.printToPDF", calculated_print_options)
        data = base64.b64decode(result['data'])

        if save_path:
            with open(save_path, 'wb') as file:
                file.write(data)
                file.close()
        return data



from contextlib import suppress
import asyncio
from concurrent.futures import ThreadPoolExecutor


class MultiTask:
    def __init__(self, func, name, args=[]):
        self.func = func
        self.args = args
        self.name = name


def has_duplicate_task_names(tasks_list: list[MultiTask]) -> tuple[bool, list[str]]:
    seen = []
    duplicates = []
    uniques = []
    for task in tasks_list:
        if task.name not in seen:
            seen.append(task.name)
            uniques.append(task)
        else:
            duplicates.append(task.name)

    return False if len(duplicates) == 0 else True, duplicates, uniques


class ProcessMultiTasks:
    def __init__(self):
        self.executor = None
        self.pending = None
        self.loop = None

    async def run_task(self, func, name, *args):
        return {'args': [*args], 'task_name': name, 'data': await self.loop.run_in_executor(self.executor, func, *args)}

    def run(self, tasks: list[MultiTask], return_when=asyncio.ALL_COMPLETED):
        has_duplicates, duplicates, uniques_tasks = has_duplicate_task_names(tasks)
        if has_duplicates:
            print(f'Detected duplicate task names: {duplicates}, running only first task of each duplicates tasks')
            tasks = uniques_tasks

        self.loop = asyncio.new_event_loop()
        self.executor = ThreadPoolExecutor(len(tasks))

        task_set = set()
        for task in tasks:
            task_set.add(self.loop.create_task(self.run_task(task.func, task.name, *task.args)))

        done, self.pending = self.loop.run_until_complete(
            asyncio.wait(task_set, return_when=return_when))

        self.cleanup()

        tasks_data = dict()
        for coroutine in done:
            tasks_data[coroutine.result().get('task_name')] = {
                'args': coroutine.result().get('args'), 'data': coroutine.result().get('data')}
        return tasks_data

    def cleanup(self) -> None:
        for p in self.pending:
            p.cancel()
            with suppress(asyncio.CancelledError):
                self.loop.run_until_complete(p)

        self.loop.close()
        self.executor = None
        self.pending = None
        self.loop = None



from nordvpn_switcher import initialize_VPN, rotate_VPN, terminate_VPN


class NordVPN:
    def __init__(self, countries=None):
        self.settings = None
        self.is_vpn_on = False

        if countries:
            self.set_countries(countries)

    def set_countries(self, countries):
        try:
            self.settings = initialize_VPN(area_input=countries)
        except Exception as e:
            print(e)

    def rotate(self, google_check=True):
        self.is_vpn_on = True
        if self.settings is not None:
            rotate_VPN(self.settings, google_check=1 if google_check else 0)
        else:
            print('Could not start vpn. Please use set_countries method before.')

    def terminate(self):
        self.settings = None
        self.is_vpn_on = False
        terminate_VPN()




from selenium.webdriver.remote.webelement import WebElement

from selenium_super.localstorage import LocalStorage


class Js:
    def __init__(self, driver):
        self.driver = driver
        self.console = Console(self.driver)
        self.window = Window(self.driver)

    def focus(self, element: WebElement) -> None:
        """ Focuses on an element. """
        self.driver.execute_script('arguments[0].focus();', element)

    def click(self, element: WebElement) -> None:
        """ Clicks on an element. """
        self.driver.execute_script('arguments[0].click();', element)

    def blur(self, element: WebElement) -> None:
        """ Clear the focus from a selected web element. """
        self.driver.execute_script('arguments[0].blur();', element)

    def scroll_into_view(self, element: WebElement) -> None:
        """ Scrolls the element into view.  """
        self.driver.execute_script("arguments[0].scrollIntoView();", element)

    def get_bounding_client_rect(self, element: WebElement) -> dict:
        """ Gets the bounding client rect of an element. """
        return self.driver.execute_script('return arguments[0].getBoundingClientRect();', element)

    def get_attribute(self, element: WebElement, attribute: str):
        """ Gets an element's attribute. """
        return self.driver.execute_script('return arguments[0].getAttribute(arguments[1]);', element, attribute)

    def set_attribute(self, element: WebElement, attribute: str, value):
        """ Sets an element's attribute. """
        return self.driver.execute_script(
            'return arguments[0].setAttribute(arguments[1], arguments[2]);', element, attribute, value
        )

    def remove_attribute(self, element: WebElement, attribute: str):
        """ Removes an element's attribute. """
        return self.driver.execute_script('return arguments[0].removeAttribute(arguments[1]);', element, attribute)

    def get_timezone_offset(self) -> int:
        """ Gets the timezone offset of the browser in minutes. """
        return self.execute_script('return new Date().getTimezoneOffset();')

    def set_coordinates(self, coordinates: tuple) -> None:
        """ Sets the geolocation for location services. """
        script = '''var latitude = arguments[0];
                    var longitude = arguments[1];
                    window.navigator.geolocation.getCurrentPosition = function(success) {
                        var position = {
                            "coords" : {
                                "latitude": latitude,
                                "longitude": longitude
                            }
                        };
                        success(position);
                }'''
        self.driver.execute_script(
            script,
            *coordinates,
        )

    def get_coordinates(self) -> tuple:
        script = '''var latitude = ''
                    var longitude = ''
                    window.navigator.geolocation.getCurrentPosition(function(pos) {
                        latitude = pos.coords.latitude;
                        longitude = pos.coords.longitude;
                    });
                    return [latitude, longitude];'''
        return self.driver.execute_script(script)

    def reset_responses(self):
        remove_old_script ="""
                    let x = document.querySelectorAll('#interceptedResponse')
                    for(let i=0;i<x.length;i++){
                        x[i].remove()
                    }
                """
        self.driver.execute_script(remove_old_script)
        
        script = """
                    (function(XHR) {
                    "use strict";

                    var element = document.createElement('div');
                    element.id = "interceptedResponse";
                    element.appendChild(document.createTextNode(""));
                    document.body.appendChild(element);

                    var open = XHR.prototype.open;
                    var send = XHR.prototype.send;

                    XHR.prototype.open = function(method, url, async, user, pass) {
                        this._url = url; // want to track the url requested
                        open.call(this, method, url, async, user, pass);
                    };

                    XHR.prototype.send = function(data) {
                        var self = this;
                        var oldOnReadyStateChange;
                        var url = this._url;

                        function onReadyStateChange() {
                        if(self.status === 200 && self.readyState == 4 /* complete */) {
                            document.getElementById("interceptedResponse").innerHTML +=
                            '{"data":' + self.responseText + '}*****';
                        }
                        if(oldOnReadyStateChange) {
                            oldOnReadyStateChange();
                        }
                        }

                        if(this.addEventListener) {
                        this.addEventListener("readystatechange", onReadyStateChange,
                            false);
                        } else {
                        oldOnReadyStateChange = this.onreadystatechange;
                        this.onreadystatechange = onReadyStateChange;
                        }
                        send.call(this, data);
                    }
                    })(XMLHttpRequest);
                """
        self.web_driver.execute_script(script)
        
class Console:
    def __init__(self, driver):
        self.driver = driver

    def log(self, data: str):
        self.driver.execute_script('console.log(arguments[0]);', data)

    def clear(self):
        self.driver.execute_script('console.clear();')


class Window:
    def __init__(self, driver):
        self.driver = driver
        self.local_storage = LocalStorage(self.driver)
        self.indexed_db = IndexedDB(self.driver)

class IndexedDB:
    def __init__(self, driver):
        self.driver = driver

    def get_database_object(self, database_name: str, object_name: str) -> dict:
        """Gets an object from a database in IndexedDB by name.

        :param database_name: Name of the database to search in.
        :param object_name: Name of the object to find.
        """
        script = '''var databaseName = arguments[0];
                    var objectName = arguments[1];
                    var callback = arguments[arguments.length - 1];
                    
                    var db_request = window.indexedDB.open(databaseName);
                    
                    db_request.onerror = function(event) {
                        callback(null);
                    };

                    db_request.onsuccess = function(event) {
                    var db = db_request.result;
                    var transaction = db.transaction(objectName);
                    var objectStore = transaction.objectStore(objectName);
                    var data_request = objectStore.getAll();
                
                    data_request.onerror = function(event) {
                        callback(null);
                    };
                
                    data_request.onsuccess = function(event) {
                        callback(data_request.result);
                    };
                };'''

        return self.driver.execute_async_script(script, database_name, object_name)

    def delete_database(self, database_name: str):
        script = '''var databaseName = arguments[0];
                    var callback = arguments[arguments.length - 1];
                    
                    try {
                        var req = indexedDB.deleteDatabase(databaseName);
                    }
                    catch (DOMException) {
                        callback(false);
                    }
                    req.onsuccess = function () {
                        callback(true);
                    };
                    req.onerror = function () {
                        callback(false);
                    };
                    req.onblocked = function () {
                        callback(false);
                    };'''
        return self.driver.execute_async_script(script, database_name)


from pathlib import Path


class Auth:
    def __init__(self, driver):
        self.driver = driver

    def load_cookies(self, cookies):
        try:
            for cookie in cookies:
                self.add_cookie(cookie)
        except Exception as e:
            print('Error loading cookies, {}'.format(e))

    def load_auth_data(self, folder_path):
        has_auth_data = Path(folder_path + 'auth.json').is_file()

        if (has_auth_data):
            auth_file = self.driver.file_utils.load_json_file(folder_path + './auth.json')
            local_storage = auth_file.get('localstorage', [])

            try:
                for key, value in local_storage.items():
                    self.local_storage.set(key, value)
            except Exception as e:
                print('Error loading local storage items, {}'.format(e))

            self.load_cookies(auth_file.get('cookies', []))

        return has_auth_data

    def save_auth_data(self, folder_path):
        obj = {}
        local_storage_items = self.local_storage.items().items()

        for key, value in local_storage_items:
            obj[key] = value

        self.file_utils.write_json_file(folder_path + 'auth.json',
                                    "w+", {
                                        'cookies': self.driver.get_cookies(),
                                        'localstorage': obj
                                    },
                                    stringify=True)

    def do_login(self, login_url, username, password, elements, fetch_url=True):
        """
        Args:
            login_url (str): login url of the service
            username (str): username / email of the user
            password (str): user password
            elements (dict(dict("by", "value"))):
                        dict with 3 keys:
                        "user_input"
                        "pass_input"
                        "login_button:

                        each key should contain a dict
                        with the following keys: (example)
                        {
                            "user_input" : {"by": By.NAME, "value" : "username"},
                            "pass_input" : {"by": By.NAME, "value" : "Password"},
                            "login_button" : {"by": By.ID, "value" : "submitbut"}
                        }
            fetch_url (bool : optional): default set to true

        Returns:
            if success -
                True
            else:
                False
        """

        success = False
        validated_user_input = self.driver.validate_element(
            elements.get('user_input', {}))
        validate_pass_input = self.driver.validate_element(
            elements.get('pass_input', {}))
        validated_login_button = self.driver.validate_element(
            elements.get('login_button', {}))

        if validate_pass_input is False or validated_user_input is False or validated_login_button is False:
            print('properties are not valid')
            return False

        try:
            if fetch_url: self.driver.get(login_url)

            user_input = self.driver.find_element(
                by=elements['user_input']['by'], value=elements['user_input']['value'])
            user_input.send_keys(username)
            user_input.send_keys('\ue00c')
            pass_input = self.driver.find_element(
                by=elements['pass_input']['by'], value=elements['pass_input']['value'])
            pass_input.send_keys(password)
            pass_input.send_keys('\ue00c')
            button = self.driver.find_element(
                by=elements['login_button']['by'], value=elements['login_button']['value'])
            button.click()
            success = True

        except Exception as e:
            print('Error while trying login to -\n{}\n[Error] - {}'.format(
                login_url, e))

        return success




import asyncio
import base64
import json
import codecs
from pathlib import Path
import time
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By

from selenium_super.process_multi_tasks import MultiTask, ProcessMultiTasks

from .file_utils import FileUtils
from .localstorage import LocalStorage

class DriverHelpers():
    def __init__(self, driver):
        print('DriverHelpers is deprecated!')
        pass
        self.driver = driver
        self.local_storage = LocalStorage(self.driver)
        self.selectors = {}
    
    def validate_element(self, element):
        return element is not None \
                and element.get('by') is not None \
                and element.get('value') is not None

    def load_cookies(self, cookies): 
        try:
            for cookie in cookies:
                self.driver.add_cookie(cookie)
        except Exception as e:
            print('Error loading cookies, {}'.format(e))
        
    def load_auth_data(self, folder_path):
        has_auth_data = Path(folder_path + 'auth.json').is_file()

        if (has_auth_data):
            auth_file = FileUtils().load_json_file(folder_path + './auth.json')
            local_storage = auth_file.get('localstorage', [])
            
            try:
                for key, value in local_storage.items():
                    self.local_storage.set(key, value)
            except Exception as e:
                print('Error loading local storage items, {}'.format(e))

            self.load_cookies(auth_file.get('cookies', []))

        return has_auth_data

    def save_auth_data(self, folder_path):
        obj = {}
        local_storage_items = self.local_storage.items().items()

        for key, value in local_storage_items:
            obj[key] = value

        FileUtils().write_json_file(folder_path + 'auth.json',
                        "w+", {
                            'cookies': self.driver.get_cookies(),
                            'localstorage': obj
                        },
                        stringify=True)

    def do_login(self, login_url, username, password, elements, fetch_url=True):
        """
        Args:
            login_url (str): login url of the service
            username (str): username / email of the user
            password (str): user password
            elements (dict(dict("by", "value"))): 
                        dict with 3 keys:
                        "user_input"
                        "pass_input"
                        "login_button:
                        
                        each key should contain a dict 
                        with the following keys: (example)
                        {
                            "user_input" : {"by": By.NAME, "value" : "username"},
                            "pass_input" : {"by": By.NAME, "value" : "Password"},
                            "login_button" : {"by": By.ID, "value" : "submitbut"}
                        }

        Returns:
            if success -
                driver
            else:
                None
        """
        validated_user_input = self.validate_element(
            elements.get('user_input', {}))
        validate_pass_input = self.validate_element(
            elements.get('pass_input', {}))
        validated_login_button = self.validate_element(
            elements.get('login_button', {}))

        if validate_pass_input is False or validated_user_input is False or validated_login_button is False:
            print('properties are not valid')
            return None

        try:
            if fetch_url: self.driver.get(login_url)

            user_input = self.driver.find_element(
                by=elements['user_input']['by'], value=elements['user_input']['value'])
            user_input.send_keys(username)
            user_input.send_keys('\ue00c')
            pass_input = self.driver.find_element(
                by=elements['pass_input']['by'], value=elements['pass_input']['value'])
            pass_input.send_keys(password)
            pass_input.send_keys('\ue00c')
            button = self.driver.find_element(
                by=elements['login_button']['by'], value=elements['login_button']['value'])
            button.click()

            return self.driver
        except Exception as e:
            print('Error while trying login to -\n{}\n[Error] - {}'.format(
                login_url, e))

    def scroll_to_bottom(self, scroll_steps=250, bottom_crop_height=1000):
        verical_ordinate = scroll_steps
        page_height = self.driver.execute_script(
            "return Math.max(document.body.scrollHeight, document.body.offsetHeight, document.documentElement.clientHeight, document.documentElement.scrollHeight, document.documentElement.offsetHeight)"
        )
        total_jumps = int((page_height - bottom_crop_height) / scroll_steps)
        for i in range(0, total_jumps):
            self.driver.execute_script(
                "window.scrollTo(window.scrollY, {});".format(
                    verical_ordinate))
            verical_ordinate += scroll_steps
            time.sleep(0.20)

    def infinite_scroll_to_bottom(self, scroll_steps=250):
        verical_ordinate = scroll_steps
        last_height = None

        while True:
            self.driver.execute_script(f"window.scrollTo(window.scrollY, {verical_ordinate});")
            verical_ordinate += scroll_steps
            height = self.driver.execute_script("return (window.innerHeight + window.scrollY)")
            if last_height != height:
                last_height = height
                time.sleep(0.20)
            else:
                break
            
    def wait_until_element_disappear(self, selector):
        self.driver.implicitly_wait(0)
        WebDriverWait(self.driver, 30).until(EC.invisibility_of_element_located((selector.get('by'), selector.get('value'))))

    def load_selectors_file(self, path):
        config = FileUtils().load_json_file(path)
        by_options = {
            "By.XPATH": By.XPATH,
            "By.NAME": By.NAME,
            "By.CLASS_NAME": By.CLASS_NAME,
            "By.CSS_SELECTOR": By.CSS_SELECTOR,
            "By.ID": By.ID,
            "By.LINK_TEXT": By.LINK_TEXT,
            "By.TAG_NAME": By.TAG_NAME
        }

        loaded_config = {}
        for key, val in config.items():
            by = by_options[val.get('by')]
            value = val.get('value')
            loaded_config[key] = {"by": by, "value": value}

        self.selectors = loaded_config
        return loaded_config

    def get_selectors(self):
        return self.selectors
    
    def get_selector_value(self, selector):
        return self.selectors.get(selector, {}).get('value', None)

    def get_elements_by_selector(self,
                                 selector_name=None,
                                 base_element=None,
                                 multiple=False,
                                 attribute=None,
                                 filter=None,
                                 xpath=None,
                                 wait=0):

        try:
            
            if xpath is None:
                by = self.selectors.get(selector_name)
                value = self.selectors.get(selector_name)

                if by is None or value is None:
                    print('No selector {} found in configuration file'.format(
                        selector_name))
                    return None
                
                by = by.get('by', None)
                value = value.get('value', None)

            else:
                by = By.XPATH
                value = xpath
                
            search_element = self.driver if base_element is None else base_element

            if not multiple:
                element = WebDriverWait(search_element, wait).until(
                    EC.presence_of_element_located((by, value)))

                if attribute is not None and element is not None:
                    return element.get_attribute(
                        attribute) if filter is None else filter(
                            element.get_attribute(attribute))
                return element

            elif multiple:
                elements = WebDriverWait(search_element, wait).until(
                    EC.presence_of_all_elements_located((by, value)))
                if attribute is not None and elements is not None:
                    attributes_data = []
                    for element in elements:
                        if filter is None:
                            attributes_data.append(
                                element.get_attribute(attribute))
                        else:
                            attributes_data.append(
                                filter(element.get_attribute(attribute)))
                    return attributes_data
                return elements
        except Exception as e:
            print(e)
            return None

    def get_args_for_elements_by_selector(self, selector):
        selector_name = selector.get('selector_name', None)
        base_element = selector.get('base_element', None)
        multiple = selector.get('multiple', None)
        attribute = selector.get('attribute', None)
        filter = selector.get('filter', None)
        xpath = selector.get('xpath', None)
        wait = selector.get('wait', None)
        return [selector_name, base_element, multiple, attribute, filter, xpath, wait]
    
    def get_multiple_elements_by_selectors(self, selectors, return_when=asyncio.ALL_COMPLETED):
        processor = ProcessMultiTasks()
        data = processor.run(tasks=[
            MultiTask(self.get_elements_by_selector, selector_args[0], args=(selector_args)) for selector_args in [self.get_args_for_elements_by_selector(selector) for selector in selectors]
        ], return_when=return_when)

        return data
    
    
    def wait_for_clickable_element(self,
                                   clickable_element,
                                   base_element=None,
                                   wait=5):
        if base_element is not None:
            WebDriverWait(base_element, wait).until(
                EC.element_to_be_clickable(clickable_element))
        else:
            WebDriverWait(self.driver, wait).until(
                EC.element_to_be_clickable(clickable_element))

    def get_file_content_chrome(self, uri):
        result = self.driver.execute_async_script("""
            var uri = arguments[0];
            var callback = arguments[1];
            var toBase64 = function(buffer){for(var r,n=new Uint8Array(buffer),t=n.length,a=new Uint8Array(4*Math.ceil(t/3)),i=new Uint8Array(64),o=0,c=0;64>c;++c)i[c]="ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/".charCodeAt(c);for(c=0;t-t%3>c;c+=3,o+=4)r=n[c]<<16|n[c+1]<<8|n[c+2],a[o]=i[r>>18],a[o+1]=i[r>>12&63],a[o+2]=i[r>>6&63],a[o+3]=i[63&r];return t%3===1?(r=n[t-1],a[o]=i[r>>2],a[o+1]=i[r<<4&63],a[o+2]=61,a[o+3]=61):t%3===2&&(r=(n[t-2]<<8)+n[t-1],a[o]=i[r>>10],a[o+1]=i[r>>4&63],a[o+2]=i[r<<2&63],a[o+3]=61),new TextDecoder("ascii").decode(a)};
            var xhr = new XMLHttpRequest();
            xhr.responseType = 'arraybuffer';
            xhr.onload = function(){ callback(toBase64(xhr.response)) };
            xhr.onerror = function(){ callback(xhr.status) };
            xhr.open('GET', uri);
            xhr.send();
            """, uri)
        if type(result) == int :
            raise Exception("Request failed with status %s" % result)
        return base64.b64decode(result)

    def save_screenshot(self, name, path, element):
        full_path = f"{path}{name}.png"
        print("Saving image in - {}".format(full_path))
        element.screenshot(full_path)
        return full_path
            
    def save_html(self, name, path):
        with codecs.open(f"{path}{name}.html", "w", "utf-8") as html:
            full_path = f"{path}{name}.html"
            print("Saving HTML in - {}".format(full_path))
            html.write(self.driver.page_source)
            return full_path
        
    def record_responses(self):
        logs = self.driver.get_log('performance')
        responses = []
        for log in logs:
            if log['message']:
                log = json.loads(log['message'])['message']
                try:
                    if ("Network.responseReceived" in log["method"] and "json" in log["params"]["response"]["mimeType"]):
                        try:
                            body = self.driver.execute_cdp_cmd('Network.getResponseBody', {'requestId': log["params"]["requestId"]})
                            log['body'] = json.loads(body['body'])
                            responses.append(log)
                        except Exception as e:
                            responses.append(log)
                except Exception as e:
                    print(e)   
                    
        return responses
                
    def save_responses(self, file_path=None):
        logs  = self.record_responses()
        json_object = {
            'logs' : logs,
        }
        if file_path: file = FileUtils().write_json_file(file_path=file_path, write_mode='w+', json_object=json_object, stringify=True, log=True)   
        return logs
    
    def click_element_by_xpath(self, xpath):
        self.driver.execute_script("""document.evaluate({}, document, null, XPathResult.FIRST_ORDERED_NODE_TYPE, null).singleNodeValue?.click()""".format(xpath))
        
    def send_devtools(self, cmd, params={}):
        resource = f"/session/{self.driver.session_id}/chromium/send_command_and_get_result"
        url = self.driver.command_executor._url + resource
        body = json.dumps({'cmd': cmd, 'params': params})
        response = self.driver.command_executor._request('POST', url, body)
        if response.get('status'):
            raise Exception(response.get('value'))
        return response.get('value')

    def get_pdf_from_html(self, print_options = {}, save_path=None):
        calculated_print_options = {
            'landscape': False,
            'displayHeaderFooter': False,
            'printBackground': True,
            'preferCSSPageSize': True,
        }
        calculated_print_options.update(print_options)
        result = self.send_devtools("Page.printToPDF", calculated_print_options)
        data = base64.b64decode(result['data'])
        if save_path:
            with open(save_path, 'wb') as file:
                file.write(data)
                file.close()
        return data
    



import json
import codecs
import os
import shutil
from .folder_utils import FolderUtils

class FileUtils:
    def write_json_file(self, file_path, write_mode, json_object, stringify=None, log=False):
        write_object = json_object
        directory = '/'.join(file_path.split('/')[:-1])
        FolderUtils().create_folder(directory)
        
        try:
            if log:
                print(f'File saved in - {file_path}')
                
            if stringify is None:
                with open(file_path, write_mode, encoding="utf8") as f:
                        json.dumps(write_object)
                        return f
            else:
                to_save = json.dumps(write_object, ensure_ascii=False, indent=4, sort_keys=True, default=str)
                with codecs.open(file_path, write_mode, encoding='utf-8') as f:
                    f.write(to_save)
                    return f
        
        except Exception as e:
            print("Failed to write to - {}\n{}".format(file_path))
            print(e)

    def load_json_file(self, file_path):
        try:
            if os.path.isfile(file_path):
                return json.load(open(file_path, encoding='utf-8'))
            return None
        except Exception as e:
            print("Failed to read file - {}".format(file_path))
            print(e)

    def write_file_from_array(self, file_path, write_mode, file_content):
        try:
            with open(file_path, write_mode) as f:
                f.writelines(file_content)
                f.close()
        except Exception as e:
            print("Failed to write to - {}\n{}".format(file_path, file_content))
            print(e)

    def load_rows_from_file(self, file_path, limit=None, row_contain=None):
        file_content = open(file_path, "r")
        if limit == None:
            return file_content
        else:
            rows = []
            i = 0
            for line in file_content:
                if i + 1 < limit:
                    clean_line = line.strip()
                    if row_contain is not None and row_contain in clean_line:
                        if line[0] == '/':
                            rows.append(clean_line[1:-1])
                        else:
                            rows.append(clean_line)
                    
                        i += 1
                else: break
            
            return rows

    def create_windows_shortcut(path, name, target='', url=True):    
        if url:
            shortcut = open(path + '/' + name + '.url', 'w')
            shortcut.write('[InternetShortcut]\n')
            shortcut.write('URL=%s' % target)
            shortcut.close()

    def remove_files_from_path(path):
        for filename in os.listdir(path):
            file_path = os.path.join(path, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                elif os.path.isdir(file_path):
                    shutil.rmtree(file_path)
            except Exception as e:
                print('Failed to delete %s. Reason: %s' % (file_path, e))


from datetime import datetime

class TimeUtils:
    def get_time_stamp(self, date:datetime=None, time_splitter=':', date_splitter='-', between_splitter=' ', show_seconds=False):
        seconds_text = ''
        if show_seconds:
            seconds_text = time_splitter + '%S'
        if date is not None:
            return date.strftime("%d{}%m{}%Y{}%H{}%M{}".format(date_splitter, date_splitter,between_splitter, time_splitter,seconds_text)) 
        return datetime.now().strftime("%d{}%m{}%Y{}%H{}%M{}".format(date_splitter, date_splitter,between_splitter, time_splitter,seconds_text)) 
        
    def get_file_time_stamp(self, date_only=False, hour_only=False, date=None):
            ts = self.get_time_stamp(time_splitter="-",between_splitter="_", date=date)
            
            if hour_only:
                return ts.split('_')[-1]
            elif date_only:
                return ts.split('_')[0]

            return ts


def get_args_for_elements_by_selector(selector):
    selector_name = selector.get('selector_name', None)
    base_element = selector.get('base_element', None)
    multiple = selector.get('multiple', None)
    attribute = selector.get('attribute', None)
    filter = selector.get('filter', None)
    xpath = selector.get('xpath', None)
    wait = selector.get('wait', None)
    return [selector_name, base_element, multiple, attribute, filter, xpath, wait]


import requests

class TelegramApi:
    def __init__(self, chat_id, bot_token) :
        self.chat_id = chat_id
        self.bot_token = bot_token

    def send_file(self, file_path, caption):
        try:
            files = {}
            files["document"] = open(file_path, "rb")
            url = f"https://api.telegram.org/bot{self.bot_token}/sendDocument"
            return requests.get(url, params={"chat_id": self.chat_id,'caption':caption}, files=files)
        except Exception as e:
            print(e)
            return None

    def send_photo(self, file_path, caption):
        try:
            files = {}
            files["photo"] = open(file_path, "rb")
            url = f"https://api.telegram.org/bot{self.bot_token}/sendPhoto"
            return requests.get(url, params={"chat_id": self.chat_id,'caption':caption}, files=files)
        except Exception as e:
            print(e)
            return None
        
    def send_message(self, message):
        try:
            return requests.post(url=f'https://api.telegram.org/bot{self.bot_token}/sendMessage',
                   data={'chat_id': self.chat_id, 'text': message})
        except Exception as e:
            print(e)
            return None
        
   


from telethon.sync import TelegramClient
from telethon.tl.functions.channels import CreateChannelRequest, CheckUsernameRequest, UpdateUsernameRequest
from telethon.tl.types import InputPeerChannel
from telethon.tl.types import DocumentAttributeVideo
from telethon import functions

class TelethonHelpers(TelegramClient):
    def __init__(self, api_id, api_hash, client_name, restrict_users=False) :
        self.client = self.get_client(api_id, api_hash, client_name, init=True)
        self.restrict_users = restrict_users
        
    def get_client(self, api_id, api_hash, client_name, init=False):
        if init == False and self.client:
            return self.client
        else:
            client = TelegramClient.__init__(client_name, api_id, api_hash)
            client.start()
            return client
    
    def get_chat_id(self, entity_name=None, entity_id=None):
        dialogs = self.client.get_dialogs()
        for dialog in dialogs:
            if (entity_name and dialog.name == entity_name) \
                or (entity_id and dialog.message.peer_id == entity_id):
                return dialog

        return None
    
    def create_group_if_not_exist(self, entity_name, group_about, messages, avatar=None):
        chat_id = self.get_chat_id(entity_name=entity_name)

        if not chat_id:
            chat_id = self.create_group(entity_name, group_about)
            if avatar: 
                upload_file_result = self.client.upload_file(file=avatar)
                try:
                    result = self.client(functions.channels.EditPhotoRequest(
                        channel=chat_id,
                        photo=upload_file_result
                    ))
                except Exception as e:
                    print(f'Error setting group avatar.\n{e}')
            
            for message in messages:
                if message.get('type') == 'message':
                    self.send_message(entity_name, message.get('message'))  
                else: 
                    self.send_file(entity_name, message.get('path'))
                            
    def send_message(self, entity_name, message):
        try:
            dialog = self.get_chat_id(entity_name=entity_name)
            
            if dialog:
                response = self.client.send_message(entity=dialog, message=message)
            else: 
                print(f'No chat found - {entity_name}')
        except Exception as e:
            print(f'Error while sending telegram message\n{e}')
  
    def send_file(self, entity_name, file, text=None):
        try:
            dialog = self.get_chat_id(entity_name=entity_name)
            if dialog:
                attributes = ()
                if '.mp4' in file:
                    attributes=(DocumentAttributeVideo(0, 0, 0),)
                response = self.client.send_file(entity=dialog, file=file, attributes=attributes, caption=text)
            else: 
                print(f'No chat found - {entity_name}')
        except Exception as e:
            print(f'Error while sending telegram message\n{e}')

    def create_group(self, group_name, group_about=None):
        createdPrivateChannel = self.client(CreateChannelRequest(group_name,group_about,megagroup=True))

        newChannelID = createdPrivateChannel.__dict__["chats"][0].__dict__["id"]
        newChannelAccessHash = createdPrivateChannel.__dict__["chats"][0].__dict__["access_hash"]
        desiredPublicUsername = "myUsernameForPublicChannel"
        checkUsernameResult = self.client(CheckUsernameRequest(InputPeerChannel(channel_id=newChannelID, access_hash=newChannelAccessHash), desiredPublicUsername))
        if(checkUsernameResult==True):
            publicChannel = self.client(UpdateUsernameRequest(InputPeerChannel(channel_id=newChannelID, access_hash=newChannelAccessHash), desiredPublicUsername))
        if self.restrict_users:
            try:
                self.client.edit_permissions(entity=group_name, view_messages=True, send_games=False, send_gifs=False, send_inline=False, send_media=False, send_messages=False, send_polls=False, send_stickers=False, change_info=False)
            except Exception as e:
                print(f'Error while trying to edit group permissions.\nError: {e}')
        return newChannelID


from setuptools import setup, find_packages 
import os
lib_folder = os.path.dirname(os.path.realpath(__file__))

try:
    # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError:
    # for pip <= 9.0.3
    from pip.req import parse_requirements


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

def load_requirements():
    reqs = parse_requirements(r'./src/selenium_super.egg-info/requirements.txt', session='install')
    return [str(ir.requirement) for ir in reqs]

setup(
    name='selenium_super',
    version='0.0.1',
    author='Eden Nahum',
    author_email='Edenik5@gmail.com',
    description='Selenium web driver with super abilities',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/edenik/selenium_super',
    project_urls = {
        "Bug Tracker": "https://github.com/edenik/selenium_super/issues"
    },
    packages=find_packages('src'),
    package_dir = {'': 'src'},
    license='MIT',
    install_requires=load_requirements(),
)


from enum import Enum
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.remote.webelement import WebElement


class Until(Enum):
    clickable = EC.element_to_be_clickable
    visible = EC.visibility_of_element_located
    visibles = EC.visibility_of_all_elements_located
    present = EC.presence_of_element_located
    presents = EC.presence_of_all_elements_located
    invisible = EC.invisibility_of_element_located


class ElementActions(Enum):
    """
    Set of supported web element actions.
    """
    click = WebElement.click
    send_keys = WebElement.send_keys
    submit = WebElement.submit
    clear = WebElement.clear


class By:
    """
    Set of supported locator strategies.
    """

    ID = "id"
    XPATH = "xpath"
    LINK_TEXT = "link text"
    PARTIAL_LINK_TEXT = "partial link text"
    NAME = "name"
    TAG_NAME = "tag name"
    CLASS_NAME = "class name"
    CSS_SELECTOR = "css selector"



import os
import glob
import shutil


class FolderUtils:
    def check_folder(self, path):
        return os.path.isdir(path)

    # Create directory if it doesn't exist to save files
    def create_folder(self, path):
        # If folder doesn't exist, then create it.
        if not self.check_folder(path):
            os.makedirs(path)

    def find_empty_dirs(self, path):
        for dirpath, dirs, files in os.walk(path):
            if not dirs and not files:
                yield dirpath

    def remove_empty_dirs(self, path):
        empty_dirs = self.find_empty_dirs(path)
        for directory in empty_dirs:
            os.rmdir(directory)

    def get_all_file_names(self, path, clean_text=None, clean_path=False, required_text=None, declined_text=None, name_parser=None):
        files = glob.glob(f"{path}*")
        if clean_text is None:
            return files
        else:
            clean_name_files = []
            for file in files:
                name = file
                has_required_text = (required_text is not None and required_text in file) or required_text is None
                has_declined_text = (declined_text is not None and declined_text in file) or declined_text is None
                if clean_text is not None:
                    for text_to_clean in clean_text:
                        name = name.replace(text_to_clean, '')
                if clean_path is True:
                    name = name.replace(path[0:-1], '').replace('\\', '')
                if name_parser is not None:
                    name = name_parser(name)
                if has_required_text and not has_declined_text:
                    clean_name_files.append(name)
            return clean_name_files

    def get_all_file_names_contains(self, path, required_text=None):
        files = glob.glob(f"{path}*")
    
        clean_name_files = []
        for file in files:
            name = file
            has_required_text = (required_text is not None and required_text in file) or (required_text is None)
            if has_required_text:
                clean_name_files.append(name)
        return clean_name_files

    def move_files_from_folder(self, source_folder, destination_folder):
        # fetch all files
        for file_name in os.listdir(source_folder):
            # construct full file path
            source = source_folder + file_name
            destination = destination_folder + file_name
            # move only files
            if os.path.isfile(source):
                shutil.move(source, destination)
                print('Moved:', file_name)
                
    def get_base_paths(self, base_path, folder_name, folder_path=os.path.realpath(__file__)):
        base_path = os.path.join(base_path,  f"../db/{folder_name}/")
        base_media_path = os.path.join(base_path,  f"../db/{folder_name}/media/")
        base_data_path = os.path.join(base_path,  f"../db/{folder_name}/data/")
        self.create_folder(base_path)
        self.create_folder(base_media_path)
        self.create_folder(base_data_path)
        folder_path = os.path.join(os.path.dirname(folder_path)) + '/'
        return base_path, base_media_path, base_data_path, folder_path